-----------------------------------------------
-- Company          :    tud
-- Author           :    schreiter
-- E-Mail           :    <email>
--
-- Filename         :   explicit.vhd
-- Project Name     :   OSM
-- Subproject Name  :   s_vhdl
-- Description      :   <short description>
--
-- Create Date      :   Fri Nov 18 10:11:24 2016
-- Last Change      :   $Date$
-- by               :   $Author$
-----------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use IEEE.math_real.all;

--SYNOPSYS translate_off

--Uncomment the following two lines to use the declarations that are 
--provided for instantiating Xilinx primitive components
--library UNISIM;
--use UNISIM.VComponents.all;

--SYNOPSYS translate_on

entity explicit is
    port ( u_i : in real;
           u_o : out real);
end entity explicit;

architecture rtl_vhdl of explicit is

  -- second order low pass, two stage ladder network R - C - R - C.
  -- output voltage = Uo(t) = U3(t) = voltage at node (3)
  -- voltage at the middle node: U2(t)
  -- Input voltage Ui(t) = U1(t) = piecewise constant
  -- in each intervall Ui(t) = const = Ui with t in [t1, t2]: 
  -- Output voltage Uo(t) = UA*exp(z1*t) + UB(z2*t) + UC
  -- with:  z1, z2 = zeros of the characteristic polynom:
  -- z1,z2 = 1/(2*R*C) * (3 +/- sqrt(5))
  -- UA = [ R*C*z2*Ui + U2(t1) -(1+R*C*z2)*U3(t1) ] / [ R*C*z1 - R*C*z2 ]
  -- UB = [ R*C*z1*Ui + U2(t1) -(1+R*C*z1)*U3(t1) ] / [ R*C*z2 - R*C*z1 ]
  -- UC = Ui


  signal u2, u3 : real := 0.0;          -- U2(t), U3(t)

  constant tau : real := 8.0e-6;        -- R*C
  constant z1 : real := (-3.0 + 2.2360679774998) / ( 2.0 * tau);  -- first zero of characteristic polynom
  constant z2 : real := (-3.0 - 2.2360679774998) / ( 2.0 * tau);  -- second zero of characteristic polynom

  signal last_u_i_time : time := 0 ns;  -- last time u_i changed
  
begin

  u_o <= u3;
  
  -- process is triggered when u_i changes.
  -- calculates new values of u1, u2 assuming u_i was constant until the recent
  -- change
  process( u_i)
    variable UA, UB, UC : real := 0.0;      -- constants within general solution
    variable u3_new, u2_new : real;       -- values of u2(t) and u3(t) at the end of
    -- the last interval, equal to the values at the beginning of the new intervall
    variable t : real;                    -- time since last u_i change in seconds
  begin
    -- t is the duration of the ending intervall
    t := real( (now - last_u_i_time) / ps ) * 1.0e-12;

    if t > 1.0e-12 then
      -- calculate u2 and u3 at the end of the ending interval, using
      -- UA, UB, UC with values from the ending interval
      u3_new := UC + UA*exp(z1*t) + UB*exp(z2*t);
      u2_new := u3_new + tau*(z1*UA*exp(z1*t) + z2*UB*exp(z2*t));
      -- caculate UA, UB, UC for the beginning interval
      UC := u_i;
      UB := (tau*z1*u_i + u2_new -(1.0+tau*z1)*u3_new ) / (tau*(z2-z1));
      UA := (tau*z2*u_i + u2_new -(1.0+tau*z2)*u3_new ) / (tau*(z1-z2));

      u2 <= u2_new;
      u3 <= u3_new;
      
      last_u_i_time <= now;
    end if;
    
  end process;

end architecture rtl_vhdl;
